from . import Matrix as Mt
import tkinter.filedialog
import time
SAVE_FILE_PATH = "save.txt"
class NeuralNet(object):
    def __init__(self, input_size, hidden_size, output_size, hidden_layer_num):
        #
        self.save_path = SAVE_FILE_PATH
        #
        self.iNodes = input_size#输入层节点数
        self.hNodes = hidden_size#每个隐藏层的节点数
        self.hLayers = hidden_layer_num#隐藏层数
        self.oNodes = output_size
        self.weights = []
        self.weights.append(Mt.Matrix(self.hNodes, self.iNodes+1))
        for i in range(self.hLayers):
            self.weights.append(Mt.Matrix(self.hNodes, self.hNodes+1))
        self.weights.append(Mt.Matrix(self.oNodes, self.hNodes+1))
        for i in range(len(self.weights)):
            self.weights[i].randomize()
        return
    #改变自身权重
    def mutate(self, mutate_rate):
        for i in range(len(self.weights)):
            self.weights[i].mutate(mutate_rate)
        return
    #获得输出层节点
    #inputs : 输入节点, 限定 : len(inputs) == self.iNodes
    #return : 返回对应的输出节点的值
    def output(self, inputs : list):
        in_mat = Mt.Matrix.fromArray(inputs)
        cur_bias = in_mat.addBias()
        for i in range(self.hLayers):
            hidden_ip = self.weights[i].dot(cur_bias)
            hidden_op = hidden_ip.activate()
            cur_bias = hidden_op.addBias()
        output_ip = self.weights[-1].dot(cur_bias)
        output = output_ip.activate()
        return output.toArray()
    #杂交, 改变权重
    #return : 返回杂交的子NeuralNet
    def crossover(self, partner):
        rst = NeuralNet(self.iNodes, self.hNodes, self.oNodes, self.hLayers)
        for i in range(len(self.weights)):
            rst.weights[i] = self.weights[i].crossover(partner.weights[i])
        return rst
    def clone(self):
        rst = NeuralNet(self.iNodes, self.hNodes, self.oNodes, self.hLayers)
        for i in range(len(self.weights)):
            rst.weights[i] = self.weights[i]
        return rst
    #从外部载入网络的权重
    def load_weights(self, weight : list):
        for i in range(len(self.weights)):
            self.weights[i] = weight[i]
        return
    #get weights
    def pull_weights(self):
        rst = []
        for i in range(len(self.weights)):
            rst.append(self.weights[i])
        return rst
    #use fixed file name
    def save(self):
        if self.save_path == "":
            return
        with open(self.save_path, "w") as save:
            save.write("{},{}\n".format(self.iNodes, self.oNodes))
            save.write("{},{}\n".format(self.hLayers, self.hNodes))
            #write weight_ih(self.hNodes, self.iNodes+1)
            for n in range(len(self.weights)):
                for i in range(self.weights[n].Rows):
                    for j in range(self.weights[n].Cols):
                        save.write("{},".format(self.weights[n].Vals[i][j]))
                    save.write("\n")
        return
    def save_as(self):
        root = tkinter.Tk()
        root.withdraw()
        self.save_path = tkinter.filedialog.askopenfilename()
        self.save()
        return
    #use fixed file name
    #assert file format
    def load(self):
        root = tkinter.Tk()
        root.withdraw()
        self.save_path = tkinter.filedialog.askopenfilename()
        if self.save_path == "":
            return
        with open(self.save_path, "r") as save:
            line = save.readline()
            vals = line.split(",")
            self.iNodes = int(vals[0])
            self.oNodes = int(vals[1])
            line = save.readline()
            vals = line.split(",")
            self.hLayers = int(vals[0])
            self.hNodes = int(vals[1])
            #valid weights
            self.weights = []
            self.weights.append(Mt.Matrix(self.hNodes, self.iNodes+1))
            for i in range(self.hLayers):
                self.weights.append(Mt.Matrix(self.hNodes, self.hNodes+1))
            self.weights.append(Mt.Matrix(self.oNodes, self.hNodes+1))
            #read weights
            for n in range(len(self.weights)):
                for i in range(self.weights[n].Rows):
                    line = save.readline()
                    vals = line.split(",")
                    for j in range(self.weights[n].Cols):
                        self.weights[n].Vals[i][j] = float(vals[j])
        return


import numpy as np
import math
import random
#激活函数
def ReLu(z):
    return z if z > 0.0 else 0.0
class Matrix(object):
    def __init__(self, row, column):
        self.Cols = column
        self.Rows = row
        self.Vals = np.zeros((row, column)) #矩阵值
        return
    #由二维数组构建
    @staticmethod
    def from2dArray(raw : np.ndarray):
        self = Matrix(0, 0)
        self.Rows = len(raw)
        self.Cols = len(raw[0])
        self.Vals = np.zeros((self.Rows, self.Cols))
        for i in range(self.Rows):
            for j in  range(self.Cols):
                self.Vals[i][j] = raw[i][j]
        return self
    #从列表获得一个单列矩阵
    @staticmethod
    def fromArray(raw : list):
        self = Matrix(len(raw), 1)
        for i in range(self.Rows):
            self.Vals[i][0] = raw[i]
        return self
    def dot(self, matrx):
        rst = Matrix(self.Rows, matrx.Cols)
        if self.Cols == matrx.Rows:
            for i in range(self.Rows):
                for j in range(matrx.Cols):
                    sum = 0.0
                    for k in range(self.Cols):
                        sum += self.Vals[i][k] * matrx.Vals[k][j]
                    rst.Vals[i][j] = sum
        return rst
    def zeroing(self):
        for i in range(self.Rows):
            for j in range(self.Cols):
                self.Vals[i][j] = 0.0
        return
    def multiply_add_assign(self, val):
        rst = Matrix(self.Rows, self.Cols)
        for i in range(self.Rows):
                for j in range(self.Cols):
                    rst.Vals[i][j] = val * self.Vals[i][j]
        return rst
    #operator"+="
    def add_and_assign(self, mtrx):
        if self.Cols == mtrx.Cols and self.Rows == mtrx.Rows:
            for i in range(self.Rows):
                for j in range(self.Cols):
                    self.Vals[i][j] += mtrx.Vals[i][j]
        return
    #随机化矩阵值
    def randomize(self):
        for i in range(self.Rows):
            for j in  range(self.Cols):
                self.Vals[i][j] = abs(0.01 * np.random.randn())
        return
    def toArray(self):
        rst = [1.0] * (self.Cols * self.Rows)
        for i in range(self.Rows):
            for j in range(self.Cols):
                rst[i * self.Cols + j] = self.Vals[i][j]
        return rst
    def addBias(self):
        rst = Matrix(self.Rows + 1, 1)
        for i in range(self.Rows):
            rst.Vals[i][0] = self.Vals[i][0]
        rst.Vals[self.Rows][0] = 1.0
        return rst
    def activate(self):
        rst = Matrix(self.Rows, self.Cols)
        for i in range(self.Rows):
             for j in range(self.Cols):
                 rst.Vals[i][j] = ReLu(self.Vals[i][j])
        return rst
    def mutate(self, rate):
        for i in range(self.Rows):
             for j in range(self.Cols):
                 if random.random() < rate:
                     #replace the randomGaussian() for lacking this function
                     self.Vals[i][j] += np.random.randn() / 3.0
                     if self.Vals[i][j] > 1.0:
                         self.Vals[i][j] = 1.0
                     elif self.Vals[i][j] < -1.0:
                         self.Vals[i][j] = -1.0
        return
    #矩阵杂交
    #限定 : shape(matrx.Vals) == shape(self.Vals)
    def crossover(self, matrx):
        rst = Matrix(self.Rows, self.Cols)
        randC = math.floor(random.random() * self.Cols)
        randR = math.floor(random.random() * self.Rows)
        for i in range(self.Rows):
            for j in range(self.Cols):
                if i < randR or (i == randR and j <= randC) :
                    rst.Vals[i][j] = self.Vals[i][j]
                else:
                    rst.Vals[i][j] = matrx.Vals[i][j]
        return rst
    def clone(self):
        rst = Matrix(self.Rows, self.Cols)
        for i in range(self.Rows):
            for j in range(self.Cols):
                rst.Vals[i][j] = self.Vals[i][j] 
        return rst
    def setVals(self, vals : list):
        for i in range(self.Rows):
            for j in range(self.Cols):
                self.Vals[i][j] = vals[i * self.Cols + j]
        return
    def __str__(self):
        str = "|"
        for i in range(self.Rows - 1):
            for j in range(self.Cols):
                str += "{},\t".format(self.Vals[i][j])
            str += "|\n|"
        for j in range(self.Cols):
            str += "{},\t".format(self.Vals[self.Rows - 1][j])
        str += "|\n"
        return str
    def rotate_90(self):
        rst = Matrix(self.Cols, self.Rows)
        for i in range(self.Rows):
            for j in range(self.Cols):
                rst.Vals[j][self.Cols - i] = self.Vals[i][j]
        return rst
    def rotate_180(self):
        rst = Matrix(self.Rows, self.Cols)
        for i in range(self.Rows):
            for j in range(self.Cols):
                rst.Vals[i][j] = self.Vals[self.Rows - 1 - i][self.Cols - 1 - j]
        return rst
    #return : the Matrix result of convolution"self * mtrx"
    #in use: self -> buffer of some picture
    #        mtrx -> conv sum(assert mtrx.Rows == mtrx.Cols)
    def conv_full(self, mtrx):
        #rotate conv_sum:
        mtrx = mtrx.rotate_180()
        #expand picture:
        pic = Matrix(self.Rows + 2 * (mtrx.Rows - 1),\
                     self.Cols + 2 * (mtrx.Cols - 1))
        s_row = mtrx.Rows - 1
        s_col = mtrx.Cols - 1
        for i in range(self.Rows):
            for j in range(self.Cols):
                pic.Vals[s_row + i][s_col + j] = self.setVals[i][j]
        #conv------------------------------------------------------
        conv_rst = Matrix(pic.Rows - mtrx.Rows + 1, pic.Cols - mtrx.Cols + 1)
        for i in range(conv_rst.Rows):
            for j in range(conv_rst.Cols):
                sum = 0.0
                for r in range(mtrx.Rows):
                    for c in range(mtrx.Cols):
                        sum += mtrx.Vals[r][c] * pic.Vals[i + r][j + c]
                conv_rst.Vals[i][j] = sum
        #----------------------------------------------------------
        return conv_rst
    def conv_same(self, mtrx):
        #rotate conv_sum:
        mtrx = mtrx.rotate_180()
        #expand picture:
        pic = Matrix(self.Rows + 2 * (mtrx.Rows - 1),\
                     self.Cols + 2 * (mtrx.Cols - 1))
        s_row = mtrx.Rows - 1
        s_col = mtrx.Cols - 1
        for i in range(self.Rows):
            for j in range(self.Cols):
                pic.Vals[s_row + i][s_col + j] = self.Vals[i][j]
        #conv------------------------------------------------------
        conv_rst = Matrix(self.Rows, self.Cols)
        shift_r = int(mtrx.Rows >> 1)
        shift_c = int(mtrx.Cols >> 1)
        for i in range(conv_rst.Rows):
            for j in range(conv_rst.Cols):
                sum = 0.0
                for r in range(mtrx.Rows):
                    for c in range(mtrx.Cols):
                        sum += mtrx.Vals[r][c] * pic.Vals[i + r + shift_r][j + c + shift_c]
                conv_rst.Vals[i][j] = sum
        #----------------------------------------------------------
        return conv_rst
    def conv_valid(self, mtrx):
        #rotate conv_sum:
        mtrx = mtrx.rotate_180()
        #conv------------------------------------------------------
        conv_rst = Matrix(self.Rows - mtrx.Rows + 1, self.Cols - mtrx.Cols + 1)
        for i in range(conv_rst.Rows):
            for j in range(conv_rst.Cols):
                sum = 0.0
                for r in range(mtrx.Rows):
                    for c in range(mtrx.Cols):
                        sum += mtrx.Vals[r][c] * self.Vals[i + r][j + c]
                conv_rst.Vals[i][j] = sum
        #----------------------------------------------------------
        return conv_rst
############################################################


from . import NeuroNetwork as NN
from . import global_vars as g
from abc import ABCMeta, abstractmethod
#abstract
#represent the role in specific game
#base class of user defined game role 
class GameRole(object):
    #assert to be default constructable
    #living_time_limit : if an individual cotinuously perform foolly for such long time, it dead
    def __init__(self, living_time_limit = 200):
        self.brain = NN.NeuralNet(g.NN_INPUT_SIZE, g.NN_HIDDEN_SIZE, g.NN_OUTPUT_SIZE,\
            g.NN_HIDDENLAYER_NUM)
        self.vision = [0.0 for i in range(g.NN_INPUT_SIZE)]
        self.descision = [0.0 for i in range(g.NN_OUTPUT_SIZE)]
        #string used to illustrate----
        self.illus_movement = ''
        self.illus_dead_reason = ''
        #-----------------------------
        self.fitness = 0.0
        self.lifetime = 0
        self.timeleft = living_time_limit
        self.timelimit = living_time_limit
        self.dead = False
        return
    @abstractmethod
    def calculate_fitness(self):
        pass
    @abstractmethod
    #look around to get self.vision(NN input)
    def look(self):
        pass
    @abstractmethod
    #draw the game role
    def show(self):
        pass
    #called after look
    #need to be overrided---------------------------------------
    def think(self):
        self.descision = self.brain.output(self.vision)
        #deal the descision, generate move-----
        #and update self.illus_movement if need

        #--------------------------------------
        return
    #-----------------------------------------------------------
    #override it to do user define game role initialize
    #take no args, use global data in overrided funtion if need
    #it can be blank
    def user_init(self):
        return
    #update user special data
    #called in Population::update()
    def user_update(self):
        return
    #just change the name'GameRole' and copy it to user subclass
    #-----------------------------------------------------------
    def clone(self):
        retval = GameRole()
        retval.brain = self.brain.clone()
        return retval
    def crossover(self, partner):
        retval = GameRole()
        retval.brain = self.brain.crossover(partner)
        return retval
    #-----------------------------------------------------------
    def mutate(self, pm):
        self.brain.mutate(pm)
        return


from . import GameRole as gr
from . import global_vars as g
import random
GA_Train_Parallel = 0
GA_Train_Sequence = 1
class Population(object):
    #indv_type : 传入自定的游戏对象(GameRole的子类)
    def __init__(self, size, pm, pc, indv_type, train_model = GA_Train_Parallel):
        self.bestIndv = indv_type() #best individual
        self.indvs = [indv_type() for i in range(size)]
        self.pm = pm
        self.pc = pc

        self.bestScore = 0 
        self.generation = 0 
        self.bestFitness = 0.0
        self.fitnessSum = 0.0

        self.tmod = train_model

        self.cur_id = 0
        self.one_down = False

        self.__interval = int(g.AG_POP_SIZE / 5) #just used to print msg
        return
    def show(self):
        if self.cur_id == 0:
            self.bestIndv.show()
        else:
            self.indvs[self.cur_id - 1].show()
        return
    def update(self): #train them in sequence order
        if self.tmod == GA_Train_Parallel:
            self._update_parallel()
        else:
            self._update_sequence()
        return
    def _update_parallel(self):
        if self._done(): #all dead generate next
            print('\n-------------<ALL Dead>--------------')
            self._generate_next()
        if not self.bestIndv.dead:
            self.bestIndv.look()
            self.bestIndv.think()
            self.bestIndv.user_update()
        for i in range(len(self.indvs)):
            if not self.indvs[i].dead:
                self.indvs[i].look()
                self.indvs[i].think()
                self.indvs[i].user_update()
        return
    def _update_sequence(self):
        if self.one_down:
            #set cur indv to next
            self.one_down = False
            if self.cur_id != len(self.indvs):
                self.cur_id += 1
            else: #all dead
                self.cur_id = 0
                print('\n-------------<ALL Dead>--------------')
                self._generate_next()
        if self.cur_id != 0:
            self.indvs[self.cur_id - 1].look()
            self.indvs[self.cur_id - 1].think()
            self.indvs[self.cur_id - 1].user_update()
        else:
            self.bestIndv.look()
            self.bestIndv.think()
            self.bestIndv.user_update()
        return
    def _calculate_fitness(self):
        self.bestIndv.calculate_fitness()
        for i in range(len(self.indvs)):
            self.indvs[i].calculate_fitness()
        return
    #all population dead
    def _done(self):
        if not self.bestIndv.dead:
            return False
        for i in range(len(self.indvs)):
            if not self.indvs[i].dead:
                return False
        return True
    #GA------------------------------------------------------------------
    def _set_best(self):
        max_fitness = 0
        max_indx = 0
        best_score = 0
        for i in range(len(self.indvs)):
            if max_fitness < self.indvs[i].fitness:
                max_fitness = self.indvs[i].fitness
                max_indx = i
        if max_fitness > self.bestIndv.fitness:
            if max_fitness > self.bestFitness:
                self.bestFitness = max_fitness
            self.bestIndv = self.indvs[max_indx].clone()
        else:
            if self.bestIndv.fitness > self.bestFitness:
                self.bestFitness = self.bestIndv.fitness
            self.bestIndv = self.bestIndv.clone()
        self.bestIndv.user_init()
        return
    def _select_parent(self):
        rand = random.uniform(0,self.fitnessSum)
        summation = 0
        for i in range(len(self.indvs)):
            summation += self.indvs[i].fitness
            if summation > rand:
                return self.indvs[i]
        return self.bestIndv
    def _mutate(self):
        for i in range(len(self.indvs)):
            self.indvs[i].mutate(self.pm)
        return
    def _generate_next(self):
        self._calculate_fitness()
        #get fitness sum:-----------------------------
        sum_rcd = self.fitnessSum
        self.fitnessSum = self.bestIndv.fitness
        for i in range(len(self.indvs)):
            self.fitnessSum += self.indvs[i].fitness
        if self.fitnessSum < sum_rcd:
            print('-------------<Degenerate>--------------')
        print('Generation<{}> :\n\tBestFitness<{}>\tFitnessSum<{}>'.\
            format(self.generation, self.bestFitness, self.fitnessSum))
        #----------------------------------------------------------------
        self._set_best()#generate new best indv
        next_generation = []
        self.generation += 1
        for i in range(len(self.indvs)):
            if i % self.__interval == 0:
                print('\tNextG<{}> new NO.{} Individual'.format(self.generation, i))
            child = self._select_parent().crossover(self._select_parent())
            child.user_init()
            child.mutate(self.pm)
            next_generation.append(child)
        self.indvs = next_generation
        return
    #--------------------------------------------------------------------
    #called after create Population obj
    def user_init(self):
        self.bestIndv.user_init()
        for i in range(len(self.indvss)):
            self.indvs[i].user_init()
        return


#redefine vals to fit in specific situation
#define NN shape-----------------------------------
NN_INPUT_SIZE = 32
NN_OUTPUT_SIZE = 4
NN_HIDDENLAYER_NUM = 2
NN_HIDDEN_SIZE = 18
#define AG coefs-----------------------------------
AG_POP_SIZE = 100#种群数
AG_POP_PM = 0.1#变异率
AG_POP_PC = 0.7#交叉率



from setuptools import setup, find_packages
setup(
    name = 'gameai_frame',\
    version = '0.1.3',\
    description = ('maybe a generitic game ai frame :)'),\
    long_description = open('E:\Python\PY_APP6\gameai_frame\README.md', 'r', encoding='UTF-8').read(),\
    author = 'Ehekatl',\
    author_email = '2072498313@qq.com',\
    maintainer = 'Ehekatl',\
    maintainer_email = '2072498313@qq.com',\
    license = 'MIT License',\
    packages = find_packages(),\
    platforms = ['all'],\
    url = 'https://github.com/Iyrian/GameAI-Frame-',\
    install_requires = ['numpy', 'tkinter'],\
    classifiers = ['Programming Language :: Python :: 3.6', 'Topic :: Software Development :: Libraries'])